import asyncio
import datetime
import json
from asyncio import exceptions
from http import HTTPStatus
from typing import Any, Callable, Optional

import aiohttp
import requests
from requests.models import Response

from malevich_coretools.abstract.abstract import *  # noqa: F403
from malevich_coretools.abstract.pipeline import Pipeline
from malevich_coretools.funcs.checks import check_profile_mode
from malevich_coretools.secondary import Config, model_from_json, show_logs_func
from malevich_coretools.secondary.const import *  # noqa: F403
from malevich_coretools.secondary.helpers import (
    show_fail_app_info,
    show_logs_flatten_func_endpoint,
)

# DocsController


def get_docs(*args, **kwargs) -> ResultIds:
    return model_from_json(send_to_core_get(DOCS(None), *args, **kwargs), ResultIds)


def get_docs_id(id: str, *args, **kwargs) -> ResultDoc:
    return model_from_json(send_to_core_get(DOCS_ID(id, None), *args, **kwargs), ResultDoc)


def get_docs_name(name: str, *args, **kwargs) -> ResultDoc:
    return model_from_json(send_to_core_get(DOCS_NAME(name, None), *args, **kwargs), ResultDoc)


def post_docs(data: DocWithName, wait: bool, *args, **kwargs) -> Alias.Id:
    return send_to_core_modify(DOCS(wait), data, *args, **kwargs)


def post_docs_id(id: str, data: DocWithName, wait: bool, *args, **kwargs) -> Alias.Id:
    return send_to_core_modify(DOCS_ID(id, wait), data, *args, **kwargs)


def delete_docs_id(id: str, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(DOCS_ID(id, wait), *args, **kwargs, is_post=False)


def delete_docs(wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(DOCS(wait), *args, **kwargs, is_post=False)

# CollectionsController


def get_collections(*args, **kwargs) -> ResultOwnAndSharedIds:
    return model_from_json(send_to_core_get(COLLECTIONS(None), *args, **kwargs), ResultOwnAndSharedIds)


def get_collections_name(name: str, operation_id: Optional[str], run_id: Optional[str], *args, **kwargs) -> ResultOwnAndSharedIds:
    return model_from_json(send_to_core_get(COLLECTIONS_IDS_NAME(name, operation_id, run_id), *args, **kwargs), ResultOwnAndSharedIds)


def get_collection_name(name: str, operation_id: Optional[str], run_id: Optional[str], offset: int, limit: int, *args, **kwargs) -> ResultCollection:
    return model_from_json(send_to_core_get(COLLECTIONS_NAME(name, operation_id, run_id, offset, limit), *args, **kwargs), ResultCollection)


def get_collections_ids_groupName(name: str, operation_id: str, run_id: str, *args, **kwargs) -> ResultIds:
    return model_from_json(send_to_core_get(COLLECTIONS_IDS_GROUP_NAME(name, operation_id, run_id), *args, **kwargs), ResultIds)


def get_collections_groupName(name: str, operation_id: str, run_id: str, *args, **kwargs) -> ResultCollections:
    return model_from_json(send_to_core_get(COLLECTIONS_GROUP_NAME(name, operation_id, run_id), *args, **kwargs), ResultCollections)


def get_collections_id(id: str, offset: int, limit: int, *args, **kwargs) -> ResultCollection:
    return model_from_json(send_to_core_get(COLLECTIONS_ID(id, offset, limit), *args, **kwargs), ResultCollection)


def post_collections(data: DocsCollection, wait: bool, *args, **kwargs) -> Alias.Id:
    return send_to_core_modify(COLLECTIONS(wait), data, *args, **kwargs)


def post_collections_id(id: str, data: DocsCollection, wait: bool, *args, **kwargs) -> Alias.Id:
    return send_to_core_modify(COLLECTIONS_ID_MODIFY(id, wait), data, *args, **kwargs)


def post_collections_id_s3(id: str, data: PostS3Settings, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(COLLECTIONS_ID_S3(id, wait), data, *args, **kwargs)


def post_collections_data(data: DocsDataCollection, wait: bool=True, *args, **kwargs) -> Alias.Id:
    return send_to_core_modify(COLLECTIONS_DATA(wait), data, *args, **kwargs)


def post_collections_id_add(id: str, data: DocsCollectionChange, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(COLLECTIONS_ID_ADD(id, wait), data, *args, **kwargs)


def post_collections_id_copy(id: str, full_copy: bool, wait: bool, *args, **kwargs) -> Alias.Id:
    return send_to_core_modify(COLLECTIONS_ID_COPY(id, full_copy, wait), *args, **kwargs)


def post_collections_id_applyScheme(id: str, data: FixScheme, wait: bool, *args, **kwargs) -> Alias.Id:
    return send_to_core_modify(COLLECTIONS_APPLY_SCHEME(id, wait), data, *args, **kwargs)


def post_collections_id_fixScheme(id: str, data: FixScheme, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(COLLECTIONS_FIX_SCHEME(id, wait), data, *args, **kwargs)


def post_collections_id_unfixScheme(id: str, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(COLLECTIONS_UNFIX_SCHEME(id, wait), *args, **kwargs)


def post_collections_metadata(id: str, data: CollectionMetadata, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(COLLECTIONS_METADATA(id, wait), data, *args, **kwargs)


def delete_collections(wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(COLLECTIONS(wait), *args, **kwargs, is_post=False)


def delete_collections_id(id: str, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(COLLECTIONS_ID_MODIFY(id, wait), *args, **kwargs, is_post=False)


def delete_collections_id_s3(key: str, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(COLLECTIONS_ID_S3(key, wait), *args, **kwargs, is_post=False)


def delete_collections_id_del(id: str, data: DocsCollectionChange, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(COLLECTIONS_ID_DEL(id, wait), data, *args, **kwargs, is_post=False)

# CollectionObjectsController


def get_collection_objects(path: Optional[str], recursive: Optional[str], *args, **kwargs) -> FilesDirs:
    return model_from_json(send_to_core_get(COLLECTION_OBJECTS_ALL_GET(path, recursive), *args, **kwargs), FilesDirs)


def get_collection_object(path: str, *args, **kwargs) -> bytes:
    return send_to_core_get(COLLECTION_OBJECTS_PATH(path, None, None), is_text=None, *args, **kwargs)


def post_collection_object_presigned_url(path: str, callback_url: Optional[str], expires_in: int, wait: bool, *args, **kwargs) -> str:
    return send_to_core_get(COLLECTION_OBJECTS_PRESIGN_PUT(path, callback_url, expires_in, wait), is_text=True, *args, **kwargs)


def get_collection_object_presigned_url(path: str, callback_url: Optional[str], expires_in: int, wait: bool, *args, **kwargs) -> str:
    return send_to_core_get(COLLECTION_OBJECTS_PRESIGN_GET(path, callback_url, expires_in, wait), is_text=True, *args, **kwargs)


def get_collections_object_presigned(signature: str, *args, **kwargs) -> bytes:
    return send_to_core_get(COLLECTION_OBJECTS_PRESIGN(signature, None), is_text=None, *args, **kwargs)


def post_collections_object(path: str, data: bytes, zip: bool, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify_raw(COLLECTION_OBJECTS_PATH(path, wait, zip), data, *args, **kwargs)


def post_collections_object_presigned(signature: str, data: bytes, zip: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify_raw(COLLECTION_OBJECTS_PRESIGN(signature, zip), data, *args, **kwargs)


def delete_collection_objects(wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(COLLECTION_OBJECTS_ALL(wait), *args, **kwargs, is_post=False)


def delete_collection_object(path: str, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(COLLECTION_OBJECTS_PATH(path, wait, None), *args, **kwargs, is_post=False)


# EndpointController


def get_endpoints(*args, **kwargs) -> Endpoints:
    return model_from_json(send_to_core_get(ENDPOINTS_ALL(None), *args, **kwargs), Endpoints)


def get_endpoint_by_hash(hash: str, *args, **kwargs) -> Endpoint:
    return model_from_json(send_to_core_get(ENDPOINTS(hash, None), *args, **kwargs), Endpoint)


def get_endpoint_run(hash: str, *args, **kwargs) -> Endpoint:
    return model_from_json(send_to_core_get(ENDPOINTS_RUN(hash), *args, **kwargs), EndpointRunInfo)


def run_endpoint(hash: str, data: Optional[EndpointOverride], with_show: bool, *args, **kwargs) -> Union[AppLogsWithResults, FlattenAppLogsWithResults, Any]:
    show_func = show_logs_flatten_func_endpoint if data is None or data.formatLogs else show_logs_func
    res = send_to_core_modify(ENDPOINTS_RUN(hash), data, with_show=with_show, show_func=show_func, *args, **kwargs)
    try:
        if data is None or data.formatLogs:
            res = FlattenAppLogsWithResults.model_validate_json(res)
        else:
            res = AppLogsWithResults.model_validate_json(res)
    except BaseException:
        pass
    return res


def create_endpoint(data: Endpoint, wait: bool, *args, **kwargs) -> str:
    return send_to_core_modify(ENDPOINTS_CREATE(wait), data, *args, **kwargs)


def update_endpoint(data: Endpoint, wait: bool, *args, **kwargs) -> str:
    return send_to_core_modify(ENDPOINTS_UPDATE(wait), data, *args, **kwargs)


def pause_endpoint(hash: str, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(ENDPOINTS_PAUSE(hash, wait), *args, **kwargs)


def resume_endpoint(hash: str, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(ENDPOINTS_RESUME(hash, wait), *args, **kwargs)


def delete_endpoints(wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(ENDPOINTS_ALL(wait), *args, **kwargs, is_post=False)


def delete_endpoint(hash: str, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(ENDPOINTS(hash, wait), *args, **kwargs, is_post=False)


# SchemeController


def get_schemes(*args, **kwargs) -> ResultOwnAndSharedIds:
    return model_from_json(send_to_core_get(SCHEMES(None), *args, **kwargs), ResultOwnAndSharedIds)


def get_schemes_id(id: str, *args, **kwargs) -> ResultScheme:
    return model_from_json(send_to_core_get(SCHEMES_ID(id, None), *args, **kwargs), ResultScheme)


def get_schemes_id_raw(id: str, *args, **kwargs) -> Alias.Json:
    return send_to_core_get(SCHEMES_ID_RAW(id), *args, **kwargs)


def get_schemes_mapping(scheme_from_id: str, scheme_to_id: str, *args, **kwargs) -> ResultMapping:
    return model_from_json(send_to_core_get(SCHEMES_MAPPING_IDS(scheme_from_id, scheme_to_id), *args, **kwargs), ResultMapping)


def post_schemes(data: SchemeWithName, wait: bool, *args, **kwargs) -> Alias.Id:
    return send_to_core_modify(SCHEMES(wait), data, *args, **kwargs)


def post_schemes_id(id: str, data: SchemeWithName, wait: bool, *args, **kwargs) -> Alias.Id:
    return send_to_core_modify(SCHEMES_ID(id, wait), data, *args, **kwargs)


def post_schemes_mapping(data: SchemesFixMapping, wait: bool, *args, **kwargs) -> Alias.Id:
    return send_to_core_modify(SCHEMES_MAPPING(wait), data, *args, **kwargs)


def delete_schemes(wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(SCHEMES(wait), *args, **kwargs, is_post=False)


def delete_schemes_id(id: str, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(SCHEMES_ID(id, wait), *args, **kwargs, is_post=False)


def delete_schemes_mapping(data: SchemesIds, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(SCHEMES_MAPPING(wait), data, *args, **kwargs, is_post=False)

# CommonController


def get_check_auth(*args, **kwargs) -> Alias.Info:
    return send_to_core_get(CHECK, *args, **kwargs, is_text=True)


def get_ping(*args, **kwargs) -> Alias.Info:
    return send_to_core_get(PING, *args, **kwargs, is_text=True)


# def get_mapping(wait: bool, *args, **kwargs) -> ResultDocMappingsFull:
#     return model_from_json(send_to_core_get(MAPPING, *args, **kwargs), ResultDocMappingsFull)


# def get_mapping_id(id: str, wait: bool, *args, **kwargs) -> ResultMapping:
#     return model_from_json(send_to_core_get(MAPPING_ID(id, None), *args, **kwargs), ResultMapping)


# def post_mapping(data: DocsAndScheme, wait: bool, *args, **kwargs) -> Alias.Info:
#     return send_to_core_modify(MAPPING(wait), data, *args, **kwargs)


# def delete_mapping_id(id: str, wait: bool, *args, **kwargs) -> Alias.Info:
#     return send_to_core_modify(MAPPING_ID(id, wait), *args, **kwargs, is_post=False)


# def delete_all(wait: bool, *args, **kwargs) -> Alias.Info:
#     return send_to_core_modify(COMMON_ALL(wait), *args, **kwargs, is_post=False)

# UserShareController


def get_share_collection_id(id: str, *args, **kwargs) -> ResultLogins:
    return model_from_json(send_to_core_get(SHARE_COLLECTION_ID(id, None), *args, **kwargs), ResultLogins)


def get_share_scheme_id(id: str, *args, **kwargs) -> ResultLogins:
    return model_from_json(send_to_core_get(SHARE_SCHEME_ID(id, None), *args, **kwargs), ResultLogins)


def get_share_userApp_id(id: str, *args, **kwargs) -> ResultLogins:
    return model_from_json(send_to_core_get(SHARE_USER_APP_ID(id, None), *args, **kwargs), ResultLogins)


def get_share_login(login: str, *args, **kwargs) -> ResultSharedForLogin:
    return model_from_json(send_to_core_get(SHARE_LOGIN(login), *args, **kwargs), ResultSharedForLogin)


def post_share_collection_id(id: str, data: SharedWithUsers, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(SHARE_COLLECTION_ID(id, wait), data, *args, **kwargs)


def post_share_scheme_id(id: str, data: SharedWithUsers, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(SHARE_SCHEME_ID(id, wait), data, *args, **kwargs)


def post_share_userApp_id(id: str, data: SharedWithUsers, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(SHARE_USER_APP_ID(id, wait), data, *args, **kwargs)


def post_share(data: Shared, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(SHARE(wait), data, *args, **kwargs)


def delete_share(data: Shared, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(SHARE(wait), data, *args, **kwargs, is_post=False)


def delete_share_all(wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(SHARE_ALL(wait), *args, **kwargs, is_post=False)

# RegistrationController


def get_register_login(login: str, *args, **kwargs) -> Alias.Info:
    return send_to_core_get(REGISTER_LOGIN(login, None), is_text=True, *args, **kwargs)


def get_register_all(*args, **kwargs) -> ResultLogins:
    return model_from_json(send_to_core_get(REGISTER_ALL, *args, **kwargs), ResultLogins)


def post_register(data: User, auth: Optional[AUTH]=None, conn_url: Optional[str]=None) -> Alias.Info:
    send_to_core_modify(REGISTER, data, auth=auth, with_auth=True, with_show=False, conn_url=conn_url)
    info = f"user {data.login} created"
    if Config.VERBOSE:
        Config.logger.info(info)
    return info


def delete_register(auth: Optional[AUTH]=None, conn_url: Optional[str]=None, *args, **kwargs) -> Alias.Info:
    send_to_core_modify(REGISTER, *args, **kwargs, auth=auth, with_show=False, is_post=False, conn_url=conn_url)
    login = auth[0] if auth is not None else Config.CORE_USERNAME
    info = f"user {login} removed"
    if Config.VERBOSE:
        Config.logger.info(info)
    return info


def delete_register_login(login: str, wait: bool, *args, **kwargs) -> Alias.Info:
    send_to_core_modify(REGISTER_LOGIN(login, wait), *args, **kwargs, with_show=False, is_post=False)
    info = f"user {login} removed"
    if Config.VERBOSE:
        Config.logger.info(info)
    return info

# UserAppsController


def get_userApps(*args, **kwargs) -> ResultOwnAndSharedIds:
    return model_from_json(send_to_core_get(USER_APPS(None), *args, **kwargs), ResultOwnAndSharedIds)


def get_userApps_realIds(*args, **kwargs) -> ResultOwnAndSharedIds:
    return model_from_json(send_to_core_get(USER_APPS_REAL_IDS, *args, **kwargs), ResultOwnAndSharedIds)


def get_userApps_mapIds(*args, **kwargs) -> ResultOwnAndSharedIdsMap:
    return model_from_json(send_to_core_get(USER_APPS_MAP_IDS, *args, **kwargs), ResultOwnAndSharedIdsMap)


def get_userApps_id(id: str, *args, **kwargs) -> UserApp:
    return model_from_json(send_to_core_get(USER_APPS_ID(id, None), *args, **kwargs), UserApp)


def get_userApps_realId(id: str, *args, **kwargs) -> UserApp:
    return model_from_json(send_to_core_get(USER_APPS_REAL_ID(id), *args, **kwargs), UserApp)


def post_userApps(data: UserApp, wait: bool, *args, **kwargs) -> Alias.Id:
    return send_to_core_modify(USER_APPS(wait), data, *args, **kwargs)


def post_userApps_id(id: str, data: UserApp, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(USER_APPS_ID(id, wait), data, *args, **kwargs)


def delete_userApps(wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(USER_APPS(wait), *args, **kwargs, is_post=False)


def delete_userApps_id(id: str, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(USER_APPS_ID(id, wait), *args, **kwargs, is_post=False)

# UserTasksController


def get_userTasks(*args, **kwargs) -> ResultIds:
    return model_from_json(send_to_core_get(USER_TASKS(None), *args, **kwargs), ResultIds)


def get_userTasks_realIds(*args, **kwargs) -> ResultIds:
    return model_from_json(send_to_core_get(USER_TASKS_REAL_IDS, *args, **kwargs), ResultIds)


def get_userTasks_mapIds(*args, **kwargs) -> ResultIdsMap:
    return model_from_json(send_to_core_get(USER_TASKS_MAP_IDS, *args, **kwargs), ResultIdsMap)


def get_userTasks_id(id: str, *args, **kwargs) -> UserTask:
    return model_from_json(send_to_core_get(USER_TASKS_ID(id, None), *args, **kwargs), UserTask)


def get_userTasks_realId(id: str, *args, **kwargs) -> UserTask:
    return model_from_json(send_to_core_get(USER_TASKS_REAL_ID(id), *args, **kwargs), UserTask)


def post_userTasks(data: UserTask, wait: bool, *args, **kwargs) -> Alias.Id:
    return send_to_core_modify(USER_TASKS(wait), data, *args, **kwargs)


def post_userTasks_id(id: str, data: UserTask, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(USER_TASKS_ID(id, wait), data, *args, **kwargs)


def delete_userTasks(wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(USER_TASKS(wait), *args, **kwargs, is_post=False)


def delete_userTasks_id(id: str, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(USER_TASKS_ID(id, wait), *args, **kwargs, is_post=False)

# UserPipelinesController


def get_userPipelines(*args, **kwargs) -> ResultIds:
    return model_from_json(send_to_core_get(USER_PIPELINES(None), *args, **kwargs), ResultIds)


def get_userPipelines_realIds(*args, **kwargs) -> ResultIds:
    return model_from_json(send_to_core_get(USER_PIPELINES_REAL_IDS, *args, **kwargs), ResultIds)


def get_userPipelines_mapIds(*args, **kwargs) -> ResultIdsMap:
    return model_from_json(send_to_core_get(USER_PIPELINES_MAP_IDS, *args, **kwargs), ResultIdsMap)


def get_userPipelines_id(id: str, *args, **kwargs) -> Pipeline:
    return model_from_json(send_to_core_get(USER_PIPELINES_ID(id, None), *args, **kwargs), Pipeline)


def get_userPipelines_realId(id: str, *args, **kwargs) -> Pipeline:
    return model_from_json(send_to_core_get(USER_PIPELINES_REAL_ID(id), *args, **kwargs), Pipeline)


def post_userPipelines(data: UserTask, wait: bool, *args, **kwargs) -> Alias.Id:
    return send_to_core_modify(USER_PIPELINES(wait), data, *args, **kwargs)


def post_userPipelines_id(id: str, data: UserTask, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(USER_PIPELINES_ID(id, wait), data, *args, **kwargs)


def delete_userPipelines(wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(USER_PIPELINES(wait), *args, **kwargs, is_post=False)


def delete_userPipelines_id(id: str, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(USER_PIPELINES_ID(id, wait), *args, **kwargs, is_post=False)

# UserCfgsController


def get_userCfgs(*args, **kwargs) -> ResultIds:
    return model_from_json(send_to_core_get(USER_CFGS(None), *args, **kwargs), ResultIds)


def get_userCfgs_realIds(*args, **kwargs) -> ResultIds:
    return model_from_json(send_to_core_get(USER_CFGS_REAL_IDS, *args, **kwargs), ResultIds)


def get_userCfgs_mapIds(*args, **kwargs) -> ResultIdsMap:
    return model_from_json(send_to_core_get(USER_CFGS_MAP_IDS, *args, **kwargs), ResultIdsMap)


def get_userCfgs_id(id: str, *args, **kwargs) -> ResultUserCfg:
    return model_from_json(send_to_core_get(USER_CFGS_ID(id, None), *args, **kwargs), ResultUserCfg)


def get_userCfgs_realId(id: str, *args, **kwargs) -> ResultUserCfg:
    return model_from_json(send_to_core_get(USER_CFGS_REAL_ID(id), *args, **kwargs), ResultUserCfg)


def post_userCfgs(data: UserCfg, wait: bool, *args, **kwargs) -> Alias.Id:
    return send_to_core_modify(USER_CFGS(wait), data, *args, **kwargs)


def post_userCfgs_id(id: str, data: UserCfg, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(USER_CFGS_ID(id, wait), data, *args, **kwargs)


def delete_userCfgs(wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(USER_CFGS(wait), *args, **kwargs, is_post=False)


def delete_userCfgs_id(id: str, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(USER_CFGS_ID(id, wait), *args, **kwargs, is_post=False)

# OperationResultsController


def get_operationResults(*args, **kwargs) -> ResultIds:
    return model_from_json(send_to_core_get(OPERATION_RESULTS(None), *args, **kwargs), ResultIds)


def get_operationResults_id(id: str, is_text=True, *args, **kwargs) -> Optional[str]:
    return send_to_core_get(OPERATION_RESULTS_ID(id, None), is_text=is_text, *args, **kwargs)


def delete_operationResults(wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(OPERATION_RESULTS(wait), *args, **kwargs, is_post=False)


def delete_operationResults_id(id: str, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(OPERATION_RESULTS_ID(id, wait), *args, **kwargs, is_post=False)

# TempRunController


def get_run_condition(id: str, *args, **kwargs) -> Condition:
    return model_from_json(send_to_core_get(TEMP_RUN_CONDITION(id), *args, **kwargs), Condition)


def get_run_activeRuns(*args, **kwargs) -> ResultIds:
    return model_from_json(send_to_core_get(TEMP_RUN_ACTIVE_RUNS, *args, **kwargs), ResultIds)


def get_run_mainTaskCfg(id: str, *args, **kwargs) -> MainTaskCfg:
    return model_from_json(send_to_core_get(TEMP_RUN_MAIN_TASK_CFG(id), *args, **kwargs), MainTaskCfg)


def get_run_operationsIds(task_id: str, cfg_id: Optional[str]=None, *args, **kwargs) -> ResultIds:
    return model_from_json(send_to_core_get(TEMP_RUN_OPERATIONS_IDS(task_id, cfg_id), *args, **kwargs), ResultIds)

# AdminController


def get_admin_runs(*args, **kwargs) -> AdminRunsInfo:
    return model_from_json(send_to_core_get(ADMIN_RUNS, *args, **kwargs), AdminRunsInfo)


def get_admin_runs_info(data: OperationOrNone, *args, **kwargs) -> Union[Alias.Json, Alias.Info]:
    return send_to_core_modify(ADMIN_RUNS_INFO, data, *args, **kwargs)


def post_admin_update_superuser(data: Superuser, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(ADMIN_SUPERUSER, data, *args, **kwargs)


def delete_admin_runs(data: AdminStopOperation, *args, **kwargs) -> Alias.Json:
    return send_to_core_modify(ADMIN_RUNS, data, is_post=False, *args, **kwargs)

# ManagerController


def get_manager_logs(data: LogsTask, with_show: bool, *args, **kwargs) -> AppLogs:
    return model_from_json(send_to_core_modify(MANAGER_LOGS, data, with_show=with_show, show_func=show_logs_func, *args, **kwargs), AppLogs)


def get_clickhouse_all(*args, **kwargs) -> str:
    return send_to_core_get(MANAGER_CLICKHOUSE_ALL, is_text=True, *args, **kwargs)


def get_clickhouse_id(id: str, *args, **kwargs) -> str:
    return send_to_core_get(MANAGER_CLICKHOUSE_ID(id), is_text=True, *args, **kwargs)


def get_manager_dagKeyValue_operationId(id: str, *args, **kwargs) -> Alias.Json:
    return send_to_core_get(MANAGER_DAG_KEY_VALUE_OPERATION_ID(id), *args, **kwargs)


def post_manager_dagKeyValue(data: KeysValues, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(MANAGER_DAG_KEY_VALUE(wait), data, *args, **kwargs)


def get_app_info(id: str, parse: bool, *args, **kwargs) -> Union[Alias.Json, AppFunctionsInfo]:
    res = send_to_core_get(MANAGER_APP_INFO(id), show_func = show_fail_app_info, *args, **kwargs)
    if parse:
        res = model_from_json(res, AppFunctionsInfo)
    return res


def get_app_info_by_real_id(id: str, parse: bool, *args, **kwargs) -> Union[Alias.Json, AppFunctionsInfo]:
    res = send_to_core_get(MANAGER_APP_INFO_REAL_ID(id), show_func = show_fail_app_info, *args, **kwargs)
    if parse:
        res = model_from_json(res, AppFunctionsInfo)
    return res


def get_image_info(data: JsonImage, parse: bool, *args, **kwargs) -> Union[Alias.Json, AppFunctionsInfo]:
    res = send_to_core_modify(MANAGER_IMAGE_INFO, data, with_auth=False, with_show=False, show_func = show_fail_app_info, *args, **kwargs)
    if parse:
        res = model_from_json(res, AppFunctionsInfo)
    return res


def get_task_schedules(data: Operation, with_show: bool, *args, **kwargs) -> Schedules:
    res = model_from_json(send_to_core_modify(MANAGER_TASK_SCHEDULES, data, with_show=False, *args, **kwargs), Schedules)
    if with_show:
        Config.logger.info(res)
    return res


def post_manager_task(data: MainTask, with_show: bool, long: bool, long_timeout: int, wait: bool, auth: Optional[AUTH], conn_url: Optional[str]=None, *args, **kwargs) -> Union[Alias.Id, AppLogs]:
    check_profile_mode(data.profileMode)
    res = send_to_core_modify(MANAGER_TASK(wait and not long), data, with_show=with_show, show_func=show_logs_func, auth=auth, conn_url=conn_url, *args, **kwargs)
    if wait and long:
        res = asyncio.run(__get_result(res, timeout=long_timeout, auth=auth))
    if not wait:
        return res
    return AppLogs.model_validate_json(res)


def post_manager_task_run(data: RunTask, with_show: bool, long: bool, long_timeout: int, wait: bool, auth: Optional[AUTH], conn_url: Optional[str]=None, *args, **kwargs) -> Optional[Union[Alias.Id, AppLogs]]:
    check_profile_mode(data.profileMode)
    res = send_to_core_modify(MANAGER_TASK_RUN(wait and not long), data, with_show=with_show, show_func=show_logs_func, auth=auth, conn_url=conn_url, *args, **kwargs)
    if wait and long:
        res = asyncio.run(__get_result(res, timeout=long_timeout, auth=auth))
    if data.withLogs or data.schedule is not None:
        if not wait:
            return res
        return AppLogs.model_validate_json(res)


def post_manager_pipeline(data: MainPipeline, with_show: bool, long: bool, long_timeout: int, wait: bool, auth: Optional[AUTH], conn_url: Optional[str]=None, *args, **kwargs) -> Union[Alias.Id, AppLogs]:
    check_profile_mode(data.profileMode)
    res = send_to_core_modify(MANAGER_PIPELINE(wait and not long), data, with_show=with_show, show_func=show_logs_func, auth=auth, conn_url=conn_url, *args, **kwargs)
    if wait and long:
        res = asyncio.run(__get_result(res, timeout=long_timeout, auth=auth))
    if not wait:
        return res
    return AppLogs.model_validate_json(res)


def post_manager_task_unschedule(data: UnscheduleOperation, with_show: bool, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(MANAGER_TASK_UNSCHEDULE(wait), data, with_show=with_show, *args, **kwargs)


def post_manager_task_stop(data: StopOperation, with_show: bool, wait: bool, *args, **kwargs) -> Optional[Union[Alias.Id, AppLogs]]:
    show_func = show_logs_func if data.withLogs else None
    res = send_to_core_modify(MANAGER_TASK_STOP(wait), data, with_show=with_show, show_func=show_func, *args, **kwargs)
    if data.withLogs:
        if not wait:
            return res
        return AppLogs.model_validate_json(res)


def post_manager_task_stop_all(data: StopOperationMany, wait: bool, *args, **kwargs) -> Alias.Json: # FIXME show
    return send_to_core_modify(MANAGER_TASK_STOP_ALL(wait), data, *args, **kwargs)


def post_manager_task_resume(data: Operation, wait: bool, *args, **kwargs) -> Alias.Empty:
    return send_to_core_modify(MANAGER_TASK_RESUME(wait), data, *args, **kwargs)


def post_manager_task_pause(data: Operation, wait: bool, *args, **kwargs) -> Alias.Empty:
    return send_to_core_modify(MANAGER_TASK_PAUSE(wait), data, *args, **kwargs)


def post_manager_app_stop(data: AppManage, wait: bool, *args, **kwargs) -> Alias.Empty:
    return send_to_core_modify(MANAGER_APP_STOP(wait), data, *args, **kwargs)


def post_manager_app_resume(data: AppManage, wait: bool, *args, **kwargs) -> Alias.Empty:
    return send_to_core_modify(MANAGER_APP_RESUME(wait), data, *args, **kwargs)


def post_manager_app_pause(data: AppManage, wait: bool, *args, **kwargs) -> Alias.Empty:
    return send_to_core_modify(MANAGER_APP_PAUSE(wait), data, *args, **kwargs)


def get_limits(*args, **kwargs) -> UserLimits:
    return model_from_json(send_to_core_get(LIMITS(True), *args, **kwargs), UserLimits)


def post_limits(data: Limits, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(LIMITS(wait), data, *args, **kwargs)


def post_user_limits(data: LimitsScope, wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(LIMITS_USER(wait), data, *args, **kwargs)


def get_handler_url(*args, **kwargs) -> str:
    return send_to_core_get(HANDLER_URL(None, True), is_text=True, *args, **kwargs)


def post_handler_url(url: Optional[str], wait: bool, *args, **kwargs) -> Alias.Info:
    return send_to_core_modify(HANDLER_URL(url, wait), *args, **kwargs)


async def kafka_send(data: KafkaMsg, *args, **kwargs) -> Union[Alias.Info, KafkaMsg]:
    result = await send_to_core_post_async(KAFKA_SEND, data, *args, **kwargs)
    try:
        return KafkaMsg.model_validate_json(result)
    except BaseException:
        return result

# BatchController


def post_batch(data: BatchOperations, *args, **kwargs) -> BatchResponses:
    return model_from_json(send_to_core_modify(BATCH, data, *args, **kwargs), BatchResponses)


##################################


async def __get_result(id: str, is_text: bool = True, check_time: float = LONG_SLEEP_TIME, timeout: int = WAIT_RESULT_TIMEOUT, auth: Optional[AUTH]=None, conn_url: Optional[str]=None) -> str:      # FIXME
    host = Config.HOST_PORT if conn_url is None else conn_url
    assert host is not None, "host port not set"
    if auth is None:
        auth = aiohttp.BasicAuth(login=Config.CORE_USERNAME, password=Config.CORE_PASSWORD, encoding='utf-8')

    timeout_deadline = None if timeout is None else datetime.datetime.now() + datetime.timedelta(0, timeout)
    while True:
        try:
            async with aiohttp.ClientSession(auth=auth, connector=aiohttp.TCPConnector(verify_ssl=False), timeout=AIOHTTP_TIMEOUT_MINI) as session:
                while True:
                    async with session.get(f"{host}{OPERATION_RESULTS_ID(id, None)}", headers=HEADERS) as response:
                        if response.ok:
                            if is_text:
                                result = await response.text()
                            else:
                                result = await response.json()
                            return result
                        else:
                            if timeout_deadline is not None:
                                now = datetime.datetime.now()
                                if now >= timeout_deadline:
                                    raise Exception("wait get result timeout")
                            await asyncio.sleep(check_time)
        except exceptions.TimeoutError:
            pass    # recreate ClientSession


async def __async_check_response(response: aiohttp.ClientResponse, show_func: Optional[Callable]=None):  # noqa: ANN202
    if not response.ok:
        if show_func is None:
            Config.logger.error(await response.text())
        else:
            show_func(await response.text(), err=True)
    response.raise_for_status()


def __check_response(path: str, response: Response, show_func: Optional[Callable]=None):  # noqa: ANN202
    if response.status_code >= 400:
        if show_func is None:
            text = response.text
            msg = f"failed: {text}" if len(text) > 0 else "failed"
            Config.logger.error(f"{path} {msg}")
        else:
            Config.logger.error(f"{path} failed")
            show_func(response.text, err=True)
    response.raise_for_status()


def send_to_core_get(path: str, with_auth=True, show_func: Optional[Callable]=None, is_text=False, auth: Optional[AUTH]=None, conn_url: Optional[str]=None) -> Optional[Union[str, bytes]]:
    host = Config.HOST_PORT if conn_url is None else conn_url
    assert host is not None, "host port not set"
    if auth is None or not with_auth:
        auth = (Config.CORE_USERNAME, Config.CORE_PASSWORD) if with_auth else None
    response = requests.get(f"{host}{path}", headers=HEADERS, auth=auth)
    __check_response(f"{host}{path}", response, show_func)
    if response.status_code == HTTPStatus.NO_CONTENT:
        return None
    if is_text is True:
        return response.text
    elif is_text is False:
        return response.json()
    else:
        return response.content


def send_to_core_modify(path: str, operation: Optional[Any] = None, with_auth: bool=True, with_show: Optional[bool]=None, show_func: Optional[Callable]=None, is_post: bool=True, auth: Optional[AUTH]=None, conn_url: Optional[str]=None) -> str:  # noqa: ANN401
    """modify: post by default, else - delete"""
    host = Config.HOST_PORT if conn_url is None else conn_url
    assert host is not None, "host port not set"
    if auth is None:
        auth = (Config.CORE_USERNAME, Config.CORE_PASSWORD)
    if operation is not None:
        operation = json.dumps(operation.model_dump())
    if is_post:
        response = requests.post(f"{host}{path}", data=operation, headers=HEADERS, auth=auth if with_auth else None)
    else:   # delete
        response = requests.delete(f"{host}{path}", data=operation, headers=HEADERS, auth=auth if with_auth else None)
    __check_response(f"{host}{path}", response, show_func=show_func)
    if response.status_code == HTTPStatus.NO_CONTENT:
        return ""
    result = response.text
    if with_show is None:
        with_show = Config.VERBOSE
    if with_show:
        if show_func is None:
            Config.logger.info(result)
        else:
            show_func(result)
    return result


def send_to_core_modify_raw(path: str, data: bytes, with_auth: bool=True, with_show: Optional[bool]=None, show_func: Optional[Callable]=None, is_post: bool=True, auth: Optional[AUTH]=None, conn_url: Optional[str]=None) -> str:  # noqa: ANN401
    """modify: post by default, else - delete"""
    host = Config.HOST_PORT if conn_url is None else conn_url
    assert host is not None, "host port not set"
    if auth is None:
        auth = (Config.CORE_USERNAME, Config.CORE_PASSWORD)
    if is_post:
        response = requests.post(f"{host}{path}", data=data, headers=HEADERS_RAW, auth=auth if with_auth else None)
    else:   # delete
        response = requests.delete(f"{host}{path}", data=data, headers=HEADERS_RAW, auth=auth if with_auth else None)
    __check_response(f"{host}{path}", response, show_func=show_func)
    if response.status_code == HTTPStatus.NO_CONTENT:
        return ""
    result = response.text
    if with_show is None:
        with_show = Config.VERBOSE
    if with_show:
        if show_func is None:
            Config.logger.info(result)
        else:
            show_func(result)
    return result


async def send_to_core_post_async(path: str, operation: Optional[str] = None, with_auth=True, with_show=True, show_func: Optional[Callable]=None, auth: Optional[AUTH]=None, conn_url: Optional[str]=None) -> str:
    host = Config.HOST_PORT if conn_url is None else conn_url
    assert host is not None, "host port not set"
    if auth is None or not with_auth:
        auth = aiohttp.BasicAuth(login=Config.CORE_USERNAME, password=Config.CORE_PASSWORD, encoding='utf-8') if with_auth else None
    if operation is not None:
        operation = json.dumps(operation.model_dump())

    async with aiohttp.ClientSession(auth=auth, connector=aiohttp.TCPConnector(verify_ssl=False), timeout=AIOHTTP_TIMEOUT) as session:
        async with session.post(f"{host}{path}", data=operation, headers=HEADERS) as response:
            await __async_check_response(response, show_func=show_func)
            result = await response.text()
    if with_show:
        if show_func is None:
            Config.logger.info(result)
        else:
            show_func(result)
    return result



from enum import IntEnum
from typing import Dict, List, Optional

from pydantic import BaseModel

from malevich_coretools.abstract import JsonImage
from malevich_coretools.abstract.abstract import (  # noqa: F401
    Restrictions,
    ScaleInfo,
    TaskComponent,
    TaskPolicy,
)


class PullCollectionPolicy(IntEnum):
    INIT = 0
    IF_NOT_EXIST = 1
    FORCE_RELOAD = 2
    FORCE_RELOAD_ALL = 3


class BaseArgument(BaseModel):
    id: Optional[str] = None                            # from - bindProcessorId
    indices: Optional[List[int]] = None                 # full if None
    # or
    collectionName: Optional[str] = None                # get id (or obj path) from cfg by it
    # or
    collectionId: Optional[str] = None                  # hardcode collection with id (or obj path)

    def validation(self) -> None:
        assert (self.id is not None) + (self.collectionName is not None) + (self.collectionId is not None) == 1, "one way of constructing the argument must be chosen"


class Argument(BaseArgument):
    group: Optional[List[BaseArgument]] = None          # for constructed dfs, sink
    # conditions: Optional[Dict[str, bool]] = None        # valid only for alternative, bindConditionId -> value, must be specified explicitly, then it will be derived from the pipeline structure

    def validation(self) -> None:
        if self.group is not None:
            assert self.id is None and self.collectionName is None and self.collectionId is None, "one way of constructing the argument must be chosen"
            for subarg in self.group:
                subarg.validation()
        else:
            assert (self.id is not None) + (self.collectionName is not None) + (self.collectionId is not None) == 1, "one way of constructing the argument must be chosen"


class AlternativeArgument(Argument):
    # alternative: Optional[List[Argument]] = None        # if set - should be only one valid argument with conditions
    pass


class AppEntity(BaseModel):
    cfg: Optional[str] = None                           # local cfg for processor/condition
    arguments: Dict[str, AlternativeArgument] = {}      # TODO or List[Argument]?
    conditions: Optional[Dict[str, bool]] = None        # condition bindId to it result

    loopArguments: Optional[Dict[str, AlternativeArgument]] = None  # other calls, TODO or List[Argument]?, problems
    loopConditions: Optional[Dict[str, bool]] = None    # condition bindId to it result for loop

    image: JsonImage
    platform: str = "base"
    platformSettings: str = ""


class Processor(AppEntity):
    processorId: str
    outputId: Optional[str] = None


class Condition(AppEntity):                             # at least one of (true, false) set
    conditionId: str


class Result(BaseModel):                                # save collection (only processor)
    name: str
    index: Optional[int] = None


class Pipeline(BaseModel):
    pipelineId: str
    processors: Dict[str, Processor] = {}               # bindProcessorId to Processor
    conditions: Dict[str, Condition] = {}               # bindConditionId to Condition
    results: Dict[str, List[Result]] = {}               # bindProcessorId to results
    pullCollectionPolicy: PullCollectionPolicy = PullCollectionPolicy.IF_NOT_EXIST



import json
import re
import sys
from typing import Any, Dict, List, Optional, Union


def numeric_version(version_str):  # noqa: ANN201
    try:
        # Split the version string by the period
        major, minor, patch = version_str.split(".")

        # Pad each part with leading zeros to make it 3 digits
        major = major.zfill(3)
        minor = minor.zfill(3)
        patch = patch.zfill(3)

        # Concatenate the padded parts
        numeric_version_str = f"{major}{minor}{patch}"

        # Convert the concatenated string to an integer
        result = int(numeric_version_str)
        # print(result)
        return result

    except ValueError:
        print("Invalid version string format. Expected format: X.X.X")
        return None


# from https://raw.githubusercontent.com/vast-ai/vast-python/master/vast.py
def parse_query(query_str: str, res: Dict = None) -> Dict:
    """
    Basically takes a query string (like the ones in the examples of commands for the search__offers function) and
    processes it into a dict of URL parameters to be sent to the server.

    :param str query_str:
    :param Dict res:
    :return Dict:
    """
    if res is None:
        res = {}
    if isinstance(query_str, list):
        query_str = " ".join(query_str)
    query_str = query_str.strip()
    opts = re.findall(
        "([a-zA-Z0-9_]+)( *[=><!]+| +(?:[lg]te?|nin|neq|eq|not ?eq|not ?in|in) )?( *)(\[[^\]]+\]|[^ ]+)?( *)",  # noqa: W605
        query_str,
    )

    # print(opts)
    # res = {}
    op_names = {
        ">=": "gte",
        ">": "gt",
        "gt": "gt",
        "gte": "gte",
        "<=": "lte",
        "<": "lt",
        "lt": "lt",
        "lte": "lte",
        "!=": "neq",
        "==": "eq",
        "=": "eq",
        "eq": "eq",
        "neq": "neq",
        "noteq": "neq",
        "not eq": "neq",
        "notin": "notin",
        "not in": "notin",
        "nin": "notin",
        "in": "in",
    }
    field_alias = {
        "cuda_vers": "cuda_max_good",
        "display_active": "gpu_display_active",
        "reliability": "reliability2",
        "dlperf_usd": "dlperf_per_dphtotal",
        "dph": "dph_total",
        "flops_usd": "flops_per_dphtotal",
    }
    field_multiplier = {
        "cpu_ram": 1000,
        "gpu_ram": 1000,
        "duration": 24.0 * 60.0 * 60.0,
    }

    fields = {
        "bw_nvlink",
        "compute_cap",
        "cpu_cores",
        "cpu_cores_effective",
        "cpu_ram",
        "cuda_max_good",
        "datacenter",
        "direct_port_count",
        "driver_version",
        "disk_bw",
        "disk_space",
        "dlperf",
        "dlperf_per_dphtotal",
        "dph_total",
        "duration",
        "external",
        "flops_per_dphtotal",
        "gpu_display_active",
        # "gpu_ram_free_min",
        "gpu_mem_bw",
        "gpu_name",
        "gpu_ram",
        "gpu_display_active",
        "has_avx",
        "host_id",
        "id",
        "inet_down",
        "inet_down_cost",
        "inet_up",
        "inet_up_cost",
        "machine_id",
        "min_bid",
        "mobo_name",
        "num_gpus",
        "pci_gen",
        "pcie_bw",
        "reliability2",
        "rentable",
        "rented",
        "storage_cost",
        "static_ip",
        "total_flops",
        "verification",
        "verified",
        "geolocation",
    }
    joined = "".join("".join(x) for x in opts)
    if joined != query_str:
        raise ValueError(
            "Unconsumed text. Did you forget to quote your query? "
            + repr(joined)
            + " != "
            + repr(query_str)
        )

    for field, op, _, value, _ in opts:
        value = value.strip(",[]")
        v = res.setdefault(field, {})
        op = op.strip()
        op_name = op_names.get(op)

        if field in field_alias:
            field = field_alias[field]

        if (field == "driver_version") and ("." in value):
            value = numeric_version(value)

        if field not in fields:
            print(
                f"Warning: Unrecognized field: {field}, see list of recognized fields.",
                file=sys.stderr,
            )
        if not op_name:
            raise ValueError(
                "Unknown operator. Did you forget to quote your query? "
                + repr(op).strip("u")
            )
        if op_name in ["in", "notin"]:
            value = [x.strip() for x in value.split(",") if x.strip()]
        if not value:
            raise ValueError(
                "Value cannot be blank. Did you forget to quote your query? "
                + repr((field, op, value))
            )
        if not field:
            raise ValueError(
                "Field cannot be blank. Did you forget to quote your query? "
                + repr((field, op, value))
            )
        if value in ["?", "*", "any"]:
            if op_name != "eq":
                raise ValueError("Wildcard only makes sense with equals.")
            if field in v:
                del v[field]
            if field in res:
                del res[field]
            continue

        if field in field_multiplier:
            value = float(value) * field_multiplier[field]

            if isinstance(value, str):
                v[op_name] = value.replace("_", " ")
            else:
                v[op_name] = value

        else:
            # print(value)
            if value == "true":
                v[op_name] = True
            elif value == "False":
                v[op_name] = False
            elif isinstance(value, str):
                v[op_name] = value.replace("_", " ")
            else:
                # v[op_name] = [v.replace('_', ' ') for v in value]
                v[op_name] = value

        res[field] = v

    return res


# from https://raw.githubusercontent.com/vast-ai/vast-python/master/vast.py
def parse(
    search_query: str,
    show_order: str = "score-",
    no_default: bool = False,
    instance_type="on-demand",
) -> Dict[str, Any]:
    """Query syntax:

            query = comparison comparison...
            comparison = field op value
            field = <name of a field>
            op = one of: <, <=, ==, !=, >=, >, in, notin
            value = <bool, int, float, etc> | 'any'
            bool: True, False

    note: to pass '>' and '<' on the command line, make sure to use quotes
    note: to encode a string query value (ie for gpu_name), replace any spaces ' ' with underscore '_'


    Examples:

            # search for somewhat reliable single RTX 3090 instances, filter out any duplicates or offers that conflict with our existing stopped instances
            ./vast search offers 'reliability > 0.98 num_gpus=1 gpu_name=RTX_3090 rented=False'

            # search for datacenter gpus with minimal compute_cap and total_flops
            ./vast search offers 'compute_cap > 610 total_flops > 5 datacenter=True'

            # search for reliable machines with at least 4 gpus, unverified, order by num_gpus, allow duplicates
            ./vast search offers 'reliability > 0.99  num_gpus>=4 verified=False rented=any' -o 'num_gpus-'


    Available fields:

                    Name				  Type	   Description

            bw_nvlink			   float	 bandwidth NVLink
            compute_cap:			int	   cuda compute capability*100  (ie:  650 for 6.5, 700 for 7.0)
            cpu_cores:			  int	   # virtual cpus
            cpu_cores_effective:	float	 # virtual cpus you get
            cpu_ram:				float	 system RAM in gigabytes
            cuda_vers:			  float	 machine max supported cuda version (based on driver version)
            datacenter:			 bool	  show only datacenter offers
            direct_port_count	   int	   open ports on host's router
            disk_bw:				float	 disk read bandwidth, in MB/s
            disk_space:			 float	 disk storage space, in GB
            dlperf:				 float	 DL-perf score  (see FAQ for explanation)
            dlperf_usd:			 float	 DL-perf/$
            dph:					float	 $/hour rental cost
            driver_version		  string	machine's nvidia driver version as 3 digit string ex. "535.86.05"
            duration:			   float	 max rental duration in days
            external:			   bool	  show external offers in addition to datacenter offers
            flops_usd:			  float	 TFLOPs/$
            geolocation:			string	Two letter country code. Works with operators =, !=, in, not in (e.g. geolocation not in [XV,XZ])
            gpu_mem_bw:			 float	 GPU memory bandwidth in GB/s
            gpu_name:			   string	GPU model name (no quotes, replace spaces with underscores, ie: RTX_3090 rather than 'RTX 3090')
            gpu_ram:				float	 GPU RAM in GB
            gpu_frac:			   float	 Ratio of GPUs in the offer to gpus in the system
            gpu_display_active:	 bool	  True if the GPU has a display attached
            has_avx:				bool	  CPU supports AVX instruction set.
            id:					 int	   instance unique ID
            inet_down:			  float	 internet download speed in Mb/s
            inet_down_cost:		 float	 internet download bandwidth cost in $/GB
            inet_up:				float	 internet upload speed in Mb/s
            inet_up_cost:		   float	 internet upload bandwidth cost in $/GB
            machine_id			  int	   machine id of instance
            min_bid:				float	 current minimum bid price in $/hr for interruptible
            num_gpus:			   int	   # of GPUs
            pci_gen:				float	 PCIE generation
            pcie_bw:				float	 PCIE bandwidth (CPU to GPU)
            reliability:			float	 machine reliability score (see FAQ for explanation)
            rentable:			   bool	  is the instance currently rentable
            rented:				 bool	  allow/disallow duplicates and potential conflicts with existing stopped instances
            storage_cost:		   float	 storage cost in $/GB/month
            static_ip:			  bool	  is the IP addr static/stable
            total_flops:			float	 total TFLOPs from all GPUs
            verified:			   bool	  is the machine verified
    """
    field_alias = {
        "cuda_vers": "cuda_max_good",
        "reliability": "reliability2",
        "dlperf_usd": "dlperf_per_dphtotal",
        "dph": "dph_total",
        "flops_usd": "flops_per_dphtotal",
    }
    try:
        if no_default:
            query = {}
        else:
            query = {
                "verified": {"eq": True},
                "external": {"eq": False},
                "rentable": {"eq": True},
                "rented": {"eq": False},
            }
            # query = {"verified": {"eq": True}, "external": {"eq": False}, "rentable": {"eq": True} }

        if search_query is not None:
            query = parse_query(search_query, query)

        order = []
        for name in show_order.split(","):
            name = name.strip()
            if not name:
                continue
            direction = "asc"
            if name.strip("-") != name:
                direction = "desc"
            field = name.strip("-")
            if field in field_alias:
                field = field_alias[field]
            order.append([field, direction])

        query["order"] = order
        query["type"] = instance_type
        # For backwards compatibility, support --type=interruptible option
        if query["type"] == "interruptible":
            query["type"] = "bid"
    except ValueError as e:
        print("Error: ", e)
        return None
    return query


def vast_settings(
    id: Optional[int] = None, query: Optional[Union[str, Dict[str, Any]]] = None, api_key: Optional[str] = None, disk: Optional[int] = None, assets: Optional[List[str]] = None
) -> str:
    """
    Platform settings for create app:\n
            by nothing - use default query inside\n
            by offer id - try use this id for app\n
            by query - use this query inside\n

    Examples:\n
            vast_settings()\n
            vast_settings(id="<id, that found with vast (vastai search offers)>")\n
            vast_settings(query={"dph": {"lt": "0.1"}})\n
            vast_settings(query="dph<0.1")\n
    """
    res = {}
    if id is not None:
        assert isinstance(id, int), "wrong id type"
        assert query is None, "id and query set"
        res["id"] = id
    elif query is not None:
        if isinstance(query, str):
            res["query"] = parse(query, no_default=True)  # run only with this options
        else:
            assert isinstance(query, dict), "wrong query type: it should be str or dict"
            res["query"] = query
    if api_key is not None:
        assert isinstance(api_key, str), "wrong api_key type"
        res["apiKey"] = api_key
    if disk is not None:
        assert isinstance(disk, int), "wrong disk type"
        res["disk"] = disk
    if assets is not None:
        res["assets"] = assets
    return json.dumps(res)



import json
import random as rand
import string
from typing import Any, Callable, Dict, Optional, Tuple, Union

from pydantic import BaseModel

from malevich_coretools.abstract.abstract import (
    Alias,
    AppLogs,
    FlattenAppLogsWithResults,
    LogsResult,
)
from malevich_coretools.secondary import Config
from malevich_coretools.secondary.kafka_utils import handle_logs

__all__ = ["to_json", "model_from_json", "rand_str", "bool_to_str", "show_logs", "show_logs_colored", "show_logs_func", "show_fail_app_info", "logs_streaming"]

__mini__delimiter = "-" * 25
__delimiter = "-" * 50
__colors = [f"\x1b[{i};20m" for i in range(31, 38)]
__color_reset = "\x1b[0m"


def to_json(data: Union[Dict[str, Any], Alias.Json], condition_and_msg: Tuple[Callable[[Dict[str, Any]], bool], str] = (lambda _: True, "")) -> str:
    if isinstance(data, Alias.Json):
        data = json.loads(data)     # json validation
        assert condition_and_msg[0](data), f"wrong data: {condition_and_msg[1]}"
        res = json.dumps(data)
    elif isinstance(data, dict):
        assert condition_and_msg[0](data), f"wrong data: {condition_and_msg[1]}"
        res = json.dumps(data)
    else:
        raise Exception("wrong type")
    return res


def model_from_json(data: Union[Dict[str, Any], Alias.Json], model: BaseModel):  # noqa: ANN201
    if isinstance(data, Alias.Json):
        data = json.loads(data)
    assert isinstance(data, dict), data
    return model.parse_obj(data)


def rand_str(size: int = 10, chars=string.ascii_letters) -> str:
    return ''.join(rand.choice(chars) for _ in range(size))


def bool_to_str(b: bool) -> str:
    return "true" if b else "false"


def __show_logs_result(res: LogsResult):  # noqa: ANN202
    if len(res.data) > 0:
        print("------- main:")
        print(res.data)
    for run_id, logs in res.logs.items():
        print(f"------- {run_id}:")
        userLogs = res.userLogs.get(run_id, "")
        if len(userLogs) > 0:
            print(userLogs)
            print("-------")
        print(logs)


def show_logs(app_logs: AppLogs, err: bool = False) -> None:  # noqa: ANN202
    show = Config.logger.error if err else Config.logger.info
    show(f"operation_id = {app_logs.operationId}")
    if app_logs.error is not None:
        show(f"error: {app_logs.error}")
        print(__delimiter)
    print("------- dag logs -------")
    print(app_logs.dagLogs)
    for app_name, app_log in app_logs.data.items():
        print(f"------- {app_name} -------")
        if len(app_log.data) == 1:
            __show_logs_result(app_log.data[0])
        else:
            for i, log_res in enumerate(app_log.data):
                print(f"------- {i}:")
                __show_logs_result(log_res)
                print(__mini__delimiter)
        print(__delimiter)


def show_logs_colored(app_logs: AppLogs, colors_dict: Optional[Dict[str, str]] = None) -> None:
    """colors_dict - should be unique for all app_logs by operation_id"""
    def format(log, color: Optional[str]) -> None:
        if color is None:
            Config.logger.warning(log)
        else:
            Config.logger.warning(color + log + __color_reset)

    def get_color(name: str) -> str:
        if colors_dict is None:
            return None
        color = colors_dict.get(name, None)
        if color is None:
            color = __colors[len(colors_dict) % len(__colors)]
            colors_dict[name] = color
        return color

    if app_logs.error is not None:
        format(f"error: {app_logs.error}", get_color("error"))
    if len(app_logs.dagLogs) > 0:
        color = get_color("dagLogs")
        for line in app_logs.dagLogs.splitlines():
            format(f"dag: {line}", color)
    for app_name, app_log in app_logs.data.items():
        color = get_color(f"${app_name}")
        for i, logs_result in enumerate(app_log.data):
            app_name_prefix = f"{app_name}${i}" if i != 0 else app_name
            if len(logs_result.data) > 0:
                for line in logs_result.data.splitlines():
                    format(f"{app_name_prefix}$main: {line}", color)
            if len(logs_result.logs) > 0:
                for run_id, logs in logs_result.logs.items():
                    user_logs = logs_result.userLogs.get(run_id, "")
                    if len(user_logs) > 0:
                        for line in user_logs.splitlines():
                            format(f"{app_name_prefix}${run_id}: {line}", color)
                    if len(logs) > 0:
                        for line in logs.splitlines():
                            format(f"{app_name_prefix}${run_id}: {line}", color)


def show_logs_func(data: str, err: bool = False):  # noqa: ANN201
    try:
        app_logs = AppLogs.model_validate_json(data)
        show_logs(app_logs, err=err)
    except BaseException:
        Config.logger.error("decode logs failed")
        show = Config.logger.error if err else Config.logger.info
        show(data)


def show_logs_flatten_func_endpoint(data: str, err: bool = False) -> None:
    show = Config.logger.error if err else Config.logger.info
    try:
        app_logs_flatten = FlattenAppLogsWithResults.model_validate_json(data)
        show(f"operation_id = {app_logs_flatten.operationId}")
        if app_logs_flatten.error is not None:
            show(f"error: {app_logs_flatten.error}")
            print(__delimiter)
        show(app_logs_flatten.logs)
    except BaseException:
        Config.logger.error("decode logs failed")
        show(data)


def show_fail_app_info(data: str, err: bool):  # noqa: ANN201
    assert err, "show only for fails"
    try:
        res = json.loads(data)
        assert "result" in res
        print(res["result"])
    except BaseException:
        Config.logger.error("decode unsuccessful app_info fail")
        print(data)


def logs_streaming(operation_id: str, kafka_host_port: Optional[str] = None, app_logs_show: Callable[[AppLogs], None] = show_logs_colored) -> None:
    colors_dict = {}
    for appLogs in handle_logs(operation_id, kafka_host_port=kafka_host_port):
        app_logs_show(appLogs, colors_dict=colors_dict)



from typing import Generator, Optional

from kafka import KafkaConsumer

from malevich_coretools.abstract import AppLogs
from malevich_coretools.secondary import Config


def __logs_topic(operation_id: str) -> str:
    return f"{operation_id}-logs"


def handle_logs(operation_id: str, kafka_host_port: Optional[str] = None) -> Generator[AppLogs, None, None]:
    if kafka_host_port is None:
        kafka_host_port = Config.KAFKA_HOST_PORT
    assert kafka_host_port is not None, "kafka_host_port not set"

    topic = __logs_topic(operation_id)
    consumer = KafkaConsumer(topic, bootstrap_servers=kafka_host_port)

    try:
        for message in consumer:
            if message.value == b'end':
                return
            try:
                logs = AppLogs.model_validate_json(message.value)
            except BaseException as ex:
                Config.logger.error(message.value)
                raise ex
            yield logs
    except KeyboardInterrupt:
        pass



from pydantic import BaseModel, Field


class VastPlatformQuery(BaseModel):
    bw_nvlink: float | None = Field(default=None, description="Bandwidth NVLink")
    compute_cap: int | None = Field(default=None, description="CUDA compute capability * 100 (ie: 650 for 6.5, 700 for 7.0)")
    cpu_cores: int | None = Field(default=None, description="Number of virtual CPU cores")
    cpu_cores_effective: float | None = Field(default=None, description="Effective number of CPU cores")
    cpu_ram: float | None = Field(default=None, description="System RAM in gigabytes")
    cuda_vers: float | None = Field(default=None, description="Machine max supported CUDA version (based on driver version)")
    datacenter: bool | None = Field(default=None, description="Show only datacenter offers")
    direct_port_count: int | None = Field(default=None, description="Open ports on host's router")
    disk_bw: float | None = Field(default=None, description="Disk read bandwidth, in MB/s")
    disk_space: float | None = Field(default=None, description="Disk storage space, in GB")
    dlperf: float | None = Field(default=None, description="DL-perf score")
    dlperf_usd: float | None = Field(default=None, description="DL-perf/$")
    dph: float | None = Field(default=None, description="$/hour rental cost")
    driver_version: str | None = Field(default=None, description="Machine's NVIDIA driver version as 3 digit string ex. '535.86.05'")
    duration: float | None = Field(default=None, description="Max rental duration in days")
    external: bool | None = Field(default=False, description="Show external offers in addition to datacenter offers")
    flops_usd: float | None = Field(default=None, description="TFLOPs/$")
    geolocation: str | None = Field(default=None, description="Two letter country code. Works with operators =, !=, in, not in (e.g. geolocation not in [XV,XZ])")
    gpu_mem_bw: float | None = Field(default=None, description="GPU memory bandwidth in GB/s")
    gpu_name: str | None = Field(default=None, description="GPU model name (no quotes, replace spaces with underscores, ie: RTX_3090 rather than 'RTX 3090')")
    gpu_ram: float | None = Field(default=None, description="GPU RAM in GB")
    gpu_frac: float | None = Field(default=None, description="Ratio of GPUs in the offer to GPUs in the system")
    gpu_display_active: bool | None = Field(default=None, description="True if the GPU has a display attached")
    has_avx: bool | None = Field(default=None, description="CPU supports AVX instruction set")
    id: int | None = Field(default=None, description="Instance unique ID")
    inet_down: float | None = Field(default=None, description="Internet download speed in Mb/s")
    inet_down_cost: float | None = Field(default=None, description="Internet download bandwidth cost in $/GB")
    inet_up: float | None = Field(default=None, description="Internet upload speed in Mb/s")
    inet_up_cost: float | None = Field(default=None, description="Internet upload bandwidth cost in $/GB")
    machine_id: int | None = Field(default=None, description="Machine ID of instance")
    min_bid: float | None = Field(default=None, description="Current minimum bid price in $/hr for interruptible")
    num_gpus: int | None = Field(default=None, description="# of GPUs")
    pci_gen: float | None = Field(default=None, description="PCIE generation")
    pcie_bw: float | None = Field(default=None, description="PCIE bandwidth")
    reliability: float | None = Field(default=None, description="Machine reliability score")
    rentable: bool | None = Field(default=True, description="Is the instance currently rentable")
    rented: bool | None = Field(default=False, description="Allow/disallow duplicates and potential conflicts with existing stopped instances")
    storage_cost: float | None = Field(default=None, description="Storage cost in $/GB/month")
    static_ip: bool | None = Field(default=None, description="Is the IP address static/stable")
    total_flops: float | None = Field(default=None, description="Total TFLOPs from all GPUs")
    verified: bool | None = Field(default=True, description="Is the machine verified")


class VastPlatformSettings(BaseModel):
    instance_id: int | None = None
    advanced_query_settings: VastPlatformQuery | None = Field(default=None, description="Configure search settings to find GPU instances on-demand")
    disk_size: int | None = Field(default=None, description="Disk storage space, in GB")
